/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl.abstr;

import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.util.TimeHelper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.RowMapper;
import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.core.namedparam.SqlParameterSource;
import org.springframework.jdbc.core.namedparam.SqlParameterSourceUtils;
import org.springframework.jdbc.support.GeneratedKeyHolder;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;
import java.time.Instant;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZoneOffset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * This class contains general purpose database operations. They should all make sense outside this project.
 */
public abstract class AbstractDAOImpl<T> {

    protected NamedParameterJdbcTemplate namedParamJdbcTemplate;
    @Autowired
    private DataSource dataSource;

    @PostConstruct
    private void init() {
        namedParamJdbcTemplate = new NamedParameterJdbcTemplate(dataSource);
    }

    /**
     * This is based on search time and has nothing to do with updating
     * @param instant
     * @return
     */
    protected String getTimestampInCopenhagenTime(Instant instant) {
        ZoneId zoneId = ZoneId.of("Europe/Copenhagen");
        ZoneOffset zoneOffset = zoneId.getRules().getOffset(instant);
        LocalDateTime localDateTime = LocalDateTime.ofInstant(instant, zoneOffset);
        return localDateTime.toString();
    }

    public abstract String getTableName();

    protected String getIdColumnName() {
        String[] tableNameParts = getTableName().split("\\.");
        if (tableNameParts.length == 1) { // If getTableName() returns "Name"
            return tableNameParts[0] + "PID";
        } else { // If getTableName() returns "ecpr2.Name"
            return tableNameParts[1] + "PID";
        }
    }

    protected abstract RowMapper<T> getDefaultRowMapper();

    public int rowCount() throws DAOException {
        final String sql = "SELECT COUNT(*) FROM " + getTableName();
        try {
            return namedParamJdbcTemplate.queryForObject(sql, new HashMap<>(), Integer.class);
        } catch (Exception e) {
            throw new DAOException("An error occurred executing count query: " + sql, e);
        }
    }

    /**
     * Careful! Deletes rows from the table. Typically only used for testing.
     */
    public void deleteRowsById(int fromId, int toId) {
        final MapSqlParameterSource parms = new MapSqlParameterSource()
                .addValue("from", fromId)
                .addValue("to", toId);
        namedParamJdbcTemplate.update("DELETE FROM " + getTableName() + " WHERE `" + getIdColumnName() + "` >= :from AND `" + getIdColumnName() + "` <= :to", parms);
    }

    protected Integer createSingleObject(String insertSql, T object) throws DAOException {
        if (object == null) {
            return null;
        }
        try {
            final GeneratedKeyHolder generatedKeyHolder = new GeneratedKeyHolder();
            ArrayList<T> a = new ArrayList<>();
            a.add(object);
            SqlParameterSource[] params = SqlParameterSourceUtils.createBatch(a.toArray());
            namedParamJdbcTemplate.update(insertSql, params[0], generatedKeyHolder);
            return generatedKeyHolder.getKey().intValue();
        } catch (Exception e) {
            throw new DAOException("An error occurred creating object: " + insertSql, e);
        }
    }
}
