/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.rest;

import dk.sds.ecpr2.dao.SearchPersonDAO;
import dk.sds.ecpr2.service.DeleteReservedIdJobScheduler;
import dk.sds.ecpr2.service.LuceneIndexJobScheduler;
import dk.sds.ecpr2.util.TimeHelper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.stereotype.Service;

import javax.servlet.ServletContext;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.sql.DataSource;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Properties;

/**
 * IsAlive servlet
 */
@WebServlet(name="isAlive", urlPatterns = {"/isAlive"})
@Service
public class IsAliveController extends AbstractServlet {
    @Autowired
    private DataSource dataSource;

    private BuildData buildData;

    @Autowired
    LuceneIndexJobScheduler luceneIndexJobScheduler;

    @Autowired
    DeleteReservedIdJobScheduler deleteReservedIdJobScheduler;

    @Autowired
    SearchPersonDAO searchPersonDAO;

    @Override
    public void doGet(HttpServletRequest request, HttpServletResponse response) throws IOException {
        isAlive(response);
    }

    public void isAlive(HttpServletResponse response) throws IOException {
        Integer status;
        String lineBreak = System.getProperty("line.separator");

        StringBuilder b = new StringBuilder();
        boolean databaseRespond = datasourceResponding();
        boolean dataIndexed = dataIndexed();
        if (databaseRespond && dataIndexed) {
            status = 200;
            b.append("Database Respond: ").append(status).append(" OK").append(lineBreak);
            b.append("Data Indexed: ").append(status).append(" OK").append(lineBreak);

        } else {
            status = 500;
            b.append("Internal Server Error").append(lineBreak);
            if (!databaseRespond) {
                b.append("Database check failed").append(lineBreak);
            } else if (!dataIndexed) {
                b.append("Data Yet Not Indexed").append(lineBreak);
            } else {
                b.append("Database check failed").append(lineBreak);
                b.append("Data Yet Not Indexed").append(lineBreak);
            }

        }

        BuildData data = getBuildData();
        b.append(lineBreak);
        b.append("Title: ").append(data.title).append(lineBreak);
        b.append("Build-Date: ").append(buildData.date).append(lineBreak);
        b.append("Build-Version: ").append(buildData.version).append(lineBreak);
        b.append("Builder: ").append(buildData.builder).append(lineBreak);
        b.append("Display time: ").append(TimeHelper.getCurrentTime()).append(lineBreak);
        b.append("Lucene indexer job: ").append(luceneIndexJobScheduler.getHealthInformation()).append(lineBreak);
        b.append("Delete Reserved Person Ids job: ").append(deleteReservedIdJobScheduler.getHealthInformation()).append(lineBreak);

        response.setStatus(status);
        response.getWriter().println(b.toString());
    }

    private boolean datasourceResponding() {
        try {
            new JdbcTemplate(dataSource).execute("SELECT 1");
        } catch (Exception e) {
            return false;
        }
        return true;
    }

    private boolean dataIndexed() {
        return searchPersonDAO.getFirstExecutionStatus();
    }

    private BuildData getBuildData() throws IOException {
        if (buildData == null) {
            InputStream stream = this.getClass().getResourceAsStream("/META-INF/MANIFEST.MF");
            if (stream != null) {
                Properties prop = new Properties();
                prop.load(new InputStreamReader(stream, "UTF-8"));
                stream.close();
                buildData = new BuildData(prop.getProperty("Implementation-Title"), prop.getProperty("Implementation-Version"), prop.getProperty("Built-Date"), prop.getProperty("Built-By"));
            }
        }
        return buildData;
    }

    private class BuildData {
        String title;
        String version;
        String date;
        String builder;

        public BuildData(String title, String version, String date, String builder) {
            this.title = title;
            this.version = version;
            this.date = date;
            this.builder = builder;
        }
    }
}
