/**
 * MIT License
 *
 * Copyright (c) 2017 Trifork
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.sds.ecpr2.dao.impl;

import dk.sds.ecpr2.*;
import dk.sds.ecpr2.dao.exceptions.DAOException;
import dk.sds.ecpr2.domain.*;
import dk.sds.ecpr2.service.ECPR2Service;
import dk.sds.ecpr2.util.ValueSets;
import org.junit.Before;
import org.junit.Ignore;
import org.junit.Test;
import org.springframework.beans.factory.annotation.Autowired;

import java.time.Instant;
import java.util.*;

import static org.junit.Assert.assertEquals;

public class SearchPersonDAOImplTest extends AbstractTest {

    @Autowired
    ECPR2Service ecpr2Service;

    @Autowired
    SearchPersonDAOImpl searchPersonDAO;

    @Test
    public void searchPerson_ByFamilyName_PersonFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Collections.singletonList(CreateFullPerson(1, "123", "Jacob", "Hansen"));
        List<PersonVO> created = createPersons(expected);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> list = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));

        // Assert
        PersonVO p = list.get(0);
        assertEquals(created.get(0).getPersonUUID(), p.getPersonUUID());
        assertEquals("Jacob", p.getNames().get(0).getGivenName());
        assertEquals("Hansen", p.getNames().get(0).getFamilyName());
        assertEquals("Soender Alle", p.getAddresses().get(0).getStreetAddressLine1());
        assertEquals("Aarhus", p.getAddresses().get(0).getCity());
        assertEquals("DK", p.getAddresses().get(0).getCountry());
        assertEquals("Dummy District", p.getAddresses().get(0).getDistrict());
        assertEquals("Dummy State", p.getAddresses().get(0).getState());
        assertEquals("88888888", p.getContacts().get(0).getValue());
        assertEquals("21/11/1990", p.getBirthDates().get(0).getDay() + "/" + p.getBirthDates().get(0).getMonth() + "/" + p.getBirthDates().get(0).getYear());
        assertEquals("male", p.getGenders().get(0).getValue());
    }

    @Test
    public void searchPerson_ByFamilyName_PersonNotFound() throws Exception {
        // Arrange
        List<PersonVO> persons = Collections.singletonList(CreateFullPerson(2, "234", "Jacob", "Hansen"));
        List<PersonVO> created = createPersons(persons);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> actual = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Jensen"));

        // Assert
        assertEquals(0, actual.size());
    }

    @Test
    public void searchPerson_ByDanishCharacters_PersonFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Collections.singletonList(CreateFullPerson(2, "234", "Jacob", "æøå"));
        List<PersonVO> created = createPersons(expected);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> actual = searchPersonDAO.searchPerson(Collections.singletonList("familyname:æøå"));

        // Assert
        assertEquals(expected.size(), actual.size());
        assertEquals(created.get(0).getPersonUUID(), actual.get(0).getPersonUUID());
    }

    @Test
    public void searchPerson_ByFamilyName_MultipleMatchesFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Arrays.asList(
                CreateFullPerson(1, "123", "Anders", "Hansen"),
                CreateFullPerson(2, "234", "Jacob", "Hansen"));
        List<PersonVO> created = createPersons(expected);
        searchPersonDAO.buildIndex(null);

        // Act
        List<PersonVO> actual = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));

        // Assert
        assertEquals(expected.size(), actual.size());
    }

    private List<PersonVO> createPersons(List<PersonVO> personlist) throws DAOException {
        List<PersonVO> resultList = new ArrayList<>();

        ServiceActorVO actorVO = new ServiceActorVO();
        RequestContextVO requestContextVO = new RequestContextVO();
        requestContextVO.setRequestTimeStamp(Instant.now());
        for (PersonVO personVO : personlist) {
            resultList.add(ecpr2Service.createPerson(personVO, actorVO, requestContextVO));
        }

        return resultList;
    }

    private List<PersonVO> updatePersons(List<PersonVO> personlist) throws DAOException {
        List<PersonVO> resultList = new ArrayList<>();

        ServiceActorVO actorVO = new ServiceActorVO();
        RequestContextVO requestContextVO = new RequestContextVO();
        requestContextVO.setRequestTimeStamp(Instant.now());
        for (PersonVO personVO : personlist) {
            resultList.add(ecpr2Service.updatePerson(personVO, actorVO, requestContextVO));
        }

        return resultList;
    }

    @Test
    public void updateIndex_UpdateSucces_SearchPersonByFamilyName() throws Exception {
        // Arrange
        List<PersonVO> initialVersion = Collections.singletonList(CreateFullPerson(1, "123", "Jens", "Hansen"));
//        List<PersonVO> updatedVersion = Collections.singletonList(CreateFullPerson(1, "123", "Jens-Peter", "Hansen"));

        List<PersonVO> created = createPersons(initialVersion);
        searchPersonDAO.buildIndex(null);
        List<PersonVO> firstSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));
        assertEquals("Jens", firstSearchResult.get(0).getNames().get(0).getGivenName());

        // Act
        created.get(0).getNames().get(0).setFamilyName(null);
        created.get(0).getNames().get(0).setGivenName(null);
        created.get(0).getNames().get(0).setUse(null);
        created.get(0).getNames().get(0).setNamePID(null);
        NameVO name = new NameVO();
        name.setGivenName("Jens-Peter");
        name.setFamilyName("Hansen");
        created.get(0).getNames().add(name);
        KeyIdentifierVO keyIdentifier = new KeyIdentifierVO();
        keyIdentifier.setValue(created.get(0).getIdentifiers().get(0).getValue());
        keyIdentifier.setOid(created.get(0).getIdentifiers().get(0).getOid());
        created.get(0).setKeyIdentifier(keyIdentifier);
        updatePersons(created);
        searchPersonDAO.buildIndex(Instant.now());
        List<PersonVO> secondSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));

        // Assert
        assertEquals(2, secondSearchResult.size()); // Merging of versions of a person into a single person doesn't happen in DAO, it happens in Handler
        assertEquals("Jens-Peter", secondSearchResult.get(1).getNames().get(1).getGivenName());
    }

    @Test
    public void deleteIndex_DeleteSucces_SearchPersonNotFound() throws Exception {
        // Arrange
        List<PersonVO> expected = Collections.singletonList(CreateFullPerson(2, "234", "Jacob", "Hansen"));
        List<PersonVO> created = createPersons(expected);
        searchPersonDAO.buildIndex(null);
        List<PersonVO> firstSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));
        assertEquals(expected.size(), firstSearchResult.size());

        // Act
        searchPersonDAO.deleteIndex(created);
        List<PersonVO> secondSearchResult = searchPersonDAO.searchPerson(Collections.singletonList("familyname:Hansen"));

        // Assert
        assertEquals(0, secondSearchResult.size());
    }


    private PersonVO CreateFullPerson(Integer personPID, String personUUID, String givenName, String familyName) {
        PersonVO person = new PersonVO();
        person.setPersonPID(personPID);
        person.setPersonUUID(personUUID);
        person.getAddresses().add(CreateAddress());
        person.getBirthDates().add(createBirthDate());
        person.getContacts().add(createContact());
        person.getNames().add(createName(givenName, familyName));
        person.getGenders().add(createGender());
        person.getModifiers().add(createModifier());

        return person;
    }

    private Modified createModifier() {
        Modified modifier = new Modified();
        modifier.setValidFrom(Calendar.getInstance());
        ModifiedBy modifiedBy = new ModifiedBy();
        modifiedBy.setRole("System");
        OrganisationType organisation = new OrganisationType();
        organisation.setAddressLine("Testvej 12");
        organisation.setType("Testorg");
        organisation.setName("Test hospital");
        OrganisationIdentifierType identifier = new OrganisationIdentifierType();
        identifier.setIdentifier("1234");
        identifier.setSource("SOR");
        organisation.setIdentifier(identifier);
        modifiedBy.setOrganisation(organisation);
        ModificatorPersonType person = new ModificatorPersonType();
        PersonIdentifierType personIdentifier = new PersonIdentifierType();
        personIdentifier.setPersonIdentifier("1111111118");
        personIdentifier.setSource("CPR");
        person.setPersonIdentifier(personIdentifier);
        person.setName("Anders And");
        person.setAuthorisationIdentifier("5678");
        modifiedBy.setPerson(person);
        modifier.setBy(modifiedBy);

        return modifier;
    }

    private AddressVO CreateAddress() {
        AddressVO address = new AddressVO();
        address.setStreetAddressLine1("Soender Alle");
        address.setCity("Aarhus");
        address.setCountry("DK");
        address.setDistrict("Dummy District");
        address.setPostalCode("8000");
        address.setPostBox("Postbox 1000");
        address.setState("Dummy State");
        address.setType(ValueSets.addressTypeValues[0]);
        address.setUse(ValueSets.addressUseValues[0]);
        return address;
    }

    private GenderVO createGender() {
        GenderVO gender = new GenderVO();
        gender.setValue(ValueSets.genderValues[0]);
        return gender;
    }

    private NameVO createName(String givenName, String familyName) {
        NameVO name = new NameVO();
        name.setGivenName(givenName);
        name.setFamilyName(familyName);
        name.setUse(ValueSets.nameUseValues[0]);
        return name;
    }

    private ContactVO createContact() {
        ContactVO contact = new ContactVO();
        contact.setSystem(ValueSets.contactSystemValues[0]);
        contact.setUse(ValueSets.contactUseValues[0]);
        contact.setValue("88888888");
        return contact;
    }

    private BirthDateVO createBirthDate() {
        BirthDateVO birthDate = new BirthDateVO();
        birthDate.setDay(21);
        birthDate.setMonth(11);
        birthDate.setYear(1990);

        return birthDate;
    }
}
